from pathlib import Path
import sys
import uuid

from .remote import (
    get_client,
    connect_pod,
    copy_file,
    create_kernel,
    execute_python,
    list_pods,
    list_jobs,
    delete_job,
    create_job,
)

from IPython.core.magic import (
    Magics,
    cell_magic,
    line_magic,
    magics_class,
    needs_local_scope,
)
from IPython.core.magic_arguments import (
    argument,
    magic_arguments,
    parse_argstring
)

CURRENT_PATH = Path(__file__).parent.resolve()


@magics_class
class ExternalKubernetesMagic(Magics):
    def __init__(self, shell):
        # You must call the parent constructor
        super().__init__(shell)
        self.pod_client = None
        self.job_client = None
        self.job_name = None
        self.pod_name = None
        self.namespace = None
        self.session = uuid.uuid1().hex

    def prepare_pod(self, result):
        if result["success"]:
            print(f"Container is running: {self.pod_name}")
            print("Transfer utilities ... ", end="")
            copy_file(
                self.pod_client,
                namespace=self.namespace,
                pod_name=self.pod_name,
                source_file=CURRENT_PATH / "create_kernel.py",
                dest_path="/tmp/",
            )
            copy_file(
                self.pod_client,
                namespace=self.namespace,
                pod_name=self.pod_name,
                source_file=CURRENT_PATH / "execute_in_kernel.py",
                dest_path="/tmp/",
            )
            print("Done.")
            # kernel is a list [kernel id, base url, token]
            self.kernel = create_kernel(
                self.pod_client,
                self.pod_name,
                self.namespace
                )
            print(self.kernel)
            print("Created kernel.")
            print("Connected.")
        else:
            # delete container?
            print(f'Failed to connect: {result["message"]}', file=sys.stderr)

    @needs_local_scope
    @line_magic("kube_exec")
    @cell_magic("kube_exec")
    @magic_arguments()
    @argument(
        "line",
        default="",
        type=str,
        nargs="?",
        # choices=[
        #     '', 'execute',
        #     'launch',
        #     'connect',
        #     'list',
        #     'destroy',
        # ],
        help="Command",
    )
    @argument(
        "-k",
        "--kubeconfig",
        default="~/.kube/config",
        type=str,
        help="Path to the kubeconfig file, default '.kube/config'",
    )
    @argument(
        "-ns",
        "--namespace",
        type=str,
        help="Kubernetes namespace",
    )
    @argument(
        "-p",
        "--pod-name",
        type=str,
        help="Name of a running pod to connect to",
    )
    @argument(
        "-j",
        "--job-name",
        type=str,
        help="Name of a running job to work with",
    )
    @argument(
        "-jp",
        "--job-prefix",
        type=str,
        default="",
        help="Job name prefix for new job (optional)",
    )
    @argument(
        "-i",
        "--image",
        type=str,
        help="Pod image (must be running Jupyter on launch)",
    )
    @argument(
        "-c",
        "--cores",
        type=int,
        default=2,
        help="Number of cores in remote executor pod, default 2 cores",
    )
    @argument(
        "-m",
        "--memory",
        type=int,
        default=2,
        help="Memory in remote executor pod, in GiB, default 2 (GiB)",
    )
    @argument(
        "-g",
        "--gpus",
        type=int,
        default=1,
        help="Number of GPUs in remote executor pod, default 1.",
    )
    @argument(
        "-gp",
        "--gpu-product",
        type=str,
        choices=[
            "NVIDIA-A100-SXM4-40GB-MIG-1g.5gb",
            "NVIDIA-A100-SXM4-40GB-MIG-3g.20gb",
            "NVIDIA-A100-SXM4-40GB",
            "NVIDIA-A100-SXM4-80GB",
            "NVIDIA-H100-80GB-HBM3",
        ],
        default="NVIDIA-A100-SXM4-40GB-MIG-1g.5gb",
        help="GPU selector for remote executor pod",
    )
    def execute(self, line="", cell="", local_ns=None):
        args = parse_argstring(self.execute, line)
        if not Path(args.kubeconfig).expanduser().is_file():
            print(
                f"""Error: Invalid kube-config file.
                No configuration found at {args.kubeconfig}""",
                file=sys.stderr,
            )
            return
        if args.line == "launch":
            self.job_name = f"{args.job_prefix}{uuid.uuid4().hex}"
            self.namespace = args.namespace
            print(f"Creating Job {self.job_name}")
            result = create_job(self.job_name, args)
            self.pod_client = result["pod_client"]
            self.job_client = result["job_client"]
            self.pod_name = result["pod_name"]
            self.prepare_pod(result)
        elif args.line == "connect":
            # connect to a running container in a job
            self.pod_name = args.pod_name
            self.namespace = args.namespace
            # check the container is running
            result = connect_pod(self.pod_name, args)
            self.pod_client = result["pod_client"]
            self.pod_name = result["pod_name"]
            self.prepare_pod(result)
        elif args.line == "destroy":
            if args.namespace:
                self.namespace = args.namespace
            if not self.job_client:
                self.job_client = get_client(args.kubeconfig, True)
            if args.job_name:
                self.job_name = args.job_name
            if (
                self.job_client is None
                or self.job_name is None
                or self.namespace is None
            ):
                print(
                    """No remote job specified.
                    Please connect first or give a name and namespace.""",
                    file=sys.stderr,
                )
                return
            print(f"Deleting job {self.job_name}")
            delete_job(self.job_client, self.job_name, self.namespace)
            self.name = None
            self.job_name = None
            # create a new session id
            self.session = uuid.uuid1().hex
        elif args.line == "pods":
            if not args.namespace:
                print("Error: Namespace required.", file=sys.stderr)
                return
            if not self.pod_client:
                self.pod_client = get_client(args.kubeconfig)
            return list_pods(self.pod_client, args.namespace)
        elif args.line == "jobs":
            if not args.namespace:
                print("Error: Namespace required.", file=sys.stderr)
                return
            if not self.job_client:
                self.job_client = get_client(args.kubeconfig, True)
            return list_jobs(self.job_client, args.namespace)
        else:
            code = cell or line
            if (
                self.pod_client is None
                or self.pod_name is None
                or self.namespace is None
            ):
                print(
                    """No remote execution container.
                    Please launch a container first.""",
                    file=sys.stderr,
                )
            result = execute_python(
                self.pod_client,
                pod_name=self.pod_name,
                namespace=self.namespace,
                kernel=self.kernel,
                session=self.session,
                code=code,
            )
            return result
