import io
import json
import pandas as pd
import pathlib
import tarfile
import sys
import time

from IPython.display import display

from kubernetes import client, config
from kubernetes.stream import stream
from kubernetes.client.exceptions import ApiException


"""
POD_STATUS = {
    'Pending',
    'Running',
    'Succeeded',
    'Failed',
    'Unknown',
}
"""


def get_client(config_path, job=False):
    kubeconfig = str(pathlib.Path(config_path).expanduser().absolute())
    config.load_kube_config(config_file=kubeconfig)
    return client.BatchV1Api() if job else client.CoreV1Api()


def define_pod_spec(image, cores, memory, num_gpu, gpu_type):
    resources_config = client.V1ResourceRequirements(
        requests={
            "cpu": cores,
            "memory": memory,
        },
        limits={
            "cpu": cores,
            "memory": memory,
            "nvidia.com/gpu": num_gpu,
        },
    )
    node_selector = {
        "nvidia.com/gpu.product": gpu_type,
    }
    container = client.V1Container(
        name="notebook", image=image, resources=resources_config
    )
    pod_spec = client.V1PodSpec(
        restart_policy="Never",
        containers=[container],
        node_selector=node_selector
    )
    return pod_spec


def define_job(name, namespace, image, cores, memory, num_gpu, gpu_type):
    metadata = client.V1ObjectMeta(labels={"remote_spawned": "true"})
    pod_spec = define_pod_spec(image, cores, memory, num_gpu, gpu_type)
    template = client.V1PodTemplateSpec(metadata=metadata, spec=pod_spec)
    job_spec = client.V1JobSpec(template=template, backoff_limit=0)
    job = client.V1Job(
        api_version="batch/v1",
        kind="Job",
        metadata=client.V1ObjectMeta(
            name=name,
            labels={"kueue.x-k8s.io/queue-name": f"{namespace}-user-queue"}
        ),
        spec=job_spec,
    )
    return job


def wait_for_pod(pod_client, name, ns, max_wait=60):
    resp = pod_client.read_namespaced_pod(name=name, namespace=ns)
    status = resp.status.phase
    i = 0
    while status == "Pending" and i < max_wait:
        print("Waiting...")
        time.sleep(1)
        resp = pod_client.read_namespaced_pod(name=name, namespace=ns)
        status = resp.status.phase
        i += 1
        for cs in resp.status.container_statuses:
            try:
                if cs.state.waiting.message:
                    print(f"  message: {cs.state.waiting.message}")
                if cs.state.waiting.reason:
                    print(f"  reason: {cs.state.waiting.reason}")
            except Exception:
                # no waiting message or reason
                pass
    if i >= max_wait:
        return "Timeout"
    return status


def wait_for_job(job_client, name, ns, max_wait=60):
    resp = job_client.read_namespaced_job(name=name, namespace=ns)
    active_status = resp.status.active if resp.status.active else 0
    ready_status = resp.status.ready if resp.status.ready else 0
    i = 0
    while active_status + ready_status != 2 and i < max_wait:
        print("Waiting...")
        time.sleep(1)
        resp = job_client.read_namespaced_job(name=name, namespace=ns)
        active_status = resp.status.active if resp.status.active else 0
        ready_status = resp.status.ready if resp.status.ready else 0
        i += 1
        for status_condition in resp.status.conditions:
            try:
                if (
                    status_condition.type == "Suspended"
                    and status_condition.status == "False"
                ):
                    print(
                        f"""message: {status_condition.message}
                        waiting for pod activation"""
                    )
            except Exception:
                pass
    if i >= max_wait:
        return "Timeout"
    return "Running" if active_status + ready_status >= 2 else "Unknown"


def get_pod_name_for_job(pod_client, job_name, namespace):
    pod_list = pod_client.list_namespaced_pod(
        namespace=namespace, label_selector=f"job-name={job_name}"
    )
    pod_name = pod_list.items[0].metadata.name
    return pod_name


def launch_job(
    job_client,
    job_name,
    job_namespace,
    container_image,
    cpu_cores=2,
    memory=2,
    num_gpu=0,
    gpu_product=None,
):
    job = define_job(
        name=job_name,
        namespace=job_namespace,
        image=container_image,
        cores=cpu_cores,
        memory=f"{memory}Gi",
        num_gpu=num_gpu,
        gpu_type=gpu_product,
    )
    job_client.create_namespaced_job(body=job, namespace=job_namespace)


def create_job(name, args):
    pod_client = get_client(args.kubeconfig, False)
    job_client = get_client(args.kubeconfig, True)
    result = {
        "success": False,
        "message": "An unknown error occurred.",
        "pod_client": pod_client,
        "job_client": job_client,
        "job_name": name,
        "pod_name": None,
    }
    try:
        launch_job(
            job_client=job_client,
            job_name=name,
            job_namespace=args.namespace,
            container_image=args.image,
            cpu_cores=args.cores,
            memory=f"{args.memory}",
            num_gpu=args.gpus,
            gpu_product=args.gpu_product,
        )
        job_status = wait_for_job(job_client, name, args.namespace)
        result["pod_name"] = get_pod_name_for_job(
            pod_client,
            name,
            args.namespace
            )
        if job_status == "Running":
            result["message"] = "Job is running."
            result["success"] = True
        elif job_status == "Unknown":
            result[
                "message"
            ] = "For some reason the state of the Job could not be obtained."
    except ApiException as exc:
        if exc.status == 403:
            print(f"Permission for namespace {args.namespace} denied.")
            result["success"] = False
            result[
                "message"
            ] = f"Permission for namespace {args.namespace} denied."
    return result


def connect_pod(pod_name, args):
    pod_client = get_client(args.kubeconfig, False)
    result = {
        "success": False,
        "message": "An unknown error occurred.",
        "pod_client": pod_client,
        "pod_name": pod_name,
    }
    try:
        status = wait_for_pod(pod_client, result["pod_name"], args.namespace)
        if status == "Running":
            result["success"] = True
            result["message"] = "Pod is running."
        elif status == "Timeout":
            result["success"] = False
            result[
                "message"
            ] = "Timed out while waiting for a running Pod in Active Job."
        elif status == "Succeeded":
            result["success"] = False
            result[
                "message"
            ] = """All containers in the Pod have terminated in success,
            and will not be restarted."""
        elif status == "Failed":
            result["success"] = False
            result[
                "message"
            ] = """All containers in the Pod have terminated,
            and at least one container has terminated in failure."""
        elif status == "Unknown":
            result["success"] = False
            result[
                "message"
            ] = "The state of the Job Pod could not be obtained."
    except ApiException as exc:
        if exc.status == 403:
            print(f"Permission for namespace {args.namespace} denied.")
            result["success"] = False
            result[
                "message"
            ] = f"Permission for namespace {args.namespace} denied."
    return result


def copy_file(
    kube_conn, namespace: str,
    pod_name: str,
    source_file: pathlib.Path,
    dest_path: str
):
    buf = io.BytesIO()
    with tarfile.open(fileobj=buf, mode="w:tar") as tar:
        tar.add(source_file,
                arcname=pathlib.Path(dest_path).joinpath(source_file.name))
    commands = [buf.getvalue()]

    # Copying file
    exec_command = ["tar", "xvf", "-", "-C", "/"]
    resp = stream(
        kube_conn.connect_get_namespaced_pod_exec,
        pod_name,
        namespace,
        command=exec_command,
        stderr=True,
        stdin=True,
        stdout=True,
        tty=False,
        _preload_content=False,
    )

    while resp.is_open():
        resp.update(timeout=1)
        if resp.peek_stdout():
            # print(f"STDOUT: {resp.read_stdout()}")
            # should be the file name
            pass
        if resp.peek_stderr():
            # do we need to close the stream?
            raise Exception(resp.read_stderr())
        if commands:
            c = commands.pop(0)
            resp.write_stdin(c)
        else:
            break
    resp.close()


def create_kernel(pod_client,
                  pod_name,
                  namespace,
                  output_file="/tmp/kernel_id"):
    exec_command = [
        "python",
        "/tmp/create_kernel.py",
        output_file,
    ]
    kernel = stream(
        pod_client.connect_get_namespaced_pod_exec,
        pod_name,
        namespace,
        command=exec_command,
        stderr=True,
        stdin=False,
        stdout=True,
        tty=False,
    )
    return kernel.split("\n")


def execute_python(pod_client, pod_name, namespace, kernel, session, code):
    exec_command = [
        "python",
        "/tmp/execute_in_kernel.py",
        kernel[0],
        kernel[1],
        kernel[2],
        session,
    ]
    resp = stream(
        pod_client.connect_get_namespaced_pod_exec,
        pod_name,
        namespace,
        command=exec_command,
        stderr=True,
        stdin=True,
        stdout=True,
        tty=False,
        _preload_content=False,
    )

    codejson = json.dumps({"code": code})
    resp.write_stdin(f"{codejson}\n")
    result = None
    while resp.is_open():
        resp.update(timeout=1)
        if resp.peek_stdout():
            text = resp.read_stdout()
            for line in text.split("\n"):
                if line:
                    try:
                        msg = json.loads(line)
                        if (
                            msg["msg_type"] == "execute_result"
                            or msg["msg_type"] == "display_data"
                        ):
                            if "data" in msg["content"]:
                                data = msg["content"]["data"]
                                display(data, raw=True)
                        elif msg["msg_type"] == "stream":
                            # text contains line breaks
                            if msg["content"]["name"] == "stderr":
                                print(msg["content"]["text"],
                                      end="",
                                      file=sys.stderr)
                            else:
                                print(msg["content"]["text"], end="")
                        elif msg["msg_type"] == "error":
                            for tb in msg["content"]["traceback"]:
                                print(tb, file=sys.stderr)
                    except Exception:
                        print(line)
        if resp.peek_stderr():
            print(resp.read_stderr(), file=sys.stderr)
    resp.close()
    return result


def delete_pod(pod_client, pod_name, namespace, max_wait=60):
    try:
        pod_client.delete_namespaced_pod(name=pod_name, namespace=namespace)
    except ApiException as exc:
        if exc.status == 403:
            print(f"Permission for namespace {namespace} denied.")
            return
    for i in range(max_wait):
        time.sleep(1)
        try:
            resp = pod_client.read_namespaced_pod(name=pod_name,
                                                  namespace=namespace)
            status = resp.status.phase
            print(f"Container status: {status}")
        except ApiException as exc:
            if exc.status == 404:
                # pod cannot be found
                print(f"Pod {pod_name} deleted in {i} seconds.")
                break
            else:
                raise
    print(
        f"""Pod {pod_name} is being deleted but has exceeded the wait
        time of {max_wait} seconds. The pod may have deleted after this wait
        time, please confirm its deletion, if the pod fails to delete
        please contact support"""
    )


def list_pods(pod_client, namespace):
    try:
        pod_list = pod_client.list_namespaced_pod(namespace)
    except ApiException as exc:
        if exc.status == 403:
            print(f"Permission for namespace {namespace} denied.")
            return
    pods = [
        [pod.metadata.name, pod.status.phase, pod.status.start_time]
        for pod in pod_list.items
    ]
    df = pd.DataFrame(
        pods,
        columns=["NAME", "STATUS", "START TIME"],
    )
    return df.set_index("NAME")


def process_job(job):
    job_completions = 0
    expected = job.spec.completions
    for c in job.status.conditions:
        if c.type == "Complete" and c.status:
            job_completions += 1
    return [
        job.metadata.name,
        f"{job_completions}/{expected}",
        job.status.start_time,
    ]


def list_jobs(job_client, namespace):
    try:
        job_list = job_client.list_namespaced_job(namespace)
    except ApiException as exc:
        if exc.status == 403:
            print(f"Permission for namespace {namespace} denied.")
            return
    df = pd.DataFrame(
        [process_job(job) for job in job_list.items],
        columns=["NAME", "STATUS", "START TIME"],
    )
    return df.set_index("NAME")


def delete_job(job_client, job_name, namespace, max_wait=60):
    try:
        resp = job_client.delete_namespaced_job(
            name=job_name,
            namespace=namespace,
            body=client.V1DeleteOptions(
                propagation_policy="Foreground", grace_period_seconds=10
            ),
        )
    except ApiException as exc:
        if exc.status == 404:
            print(f"Job {job_name} could not be found.")
            return
        if exc.status == 403:
            print(f"Permission for namespace {namespace} denied.")
            return
    for i in range(max_wait):
        time.sleep(1)
        try:
            resp = job_client.read_namespaced_job(name=job_name,
                                                  namespace=namespace)
            print(f"Status active: {resp.status.active}")
        except ApiException as exc:
            if exc.status == 404:
                print(f"Job {job_name} deleted in {i} seconds.")
                return
            else:
                raise

    print(
        f"""Job {job_name} is being deleted but has exceeded the wait time of
        {max_wait} seconds. The job may have deleted after this wait time,
        please confirm its deletion,
        if the job fails to delete please contact support"""
    )
